
extern struct BuiltinFileBlob bfb_4DOS_COM;
extern struct BuiltinFileBlob bfb_4DOS_HLP;
extern struct BuiltinFileBlob bfb_4HELP_EXE;
extern struct BuiltinFileBlob bfb_BATCOMP_EXE;
extern struct BuiltinFileBlob bfb_EXAMPLES_BTM;
extern struct BuiltinFileBlob bfb_LICENSE_TXT;
extern struct BuiltinFileBlob bfb_OPTION_EXE;

